import numpy as np

from abc import ABC

from .loading.loading import Loading
from .statistics.statistics import Statistics
from ..settings.settings import Settings
from ...common.probability import ProbabilityFunctions


class BaseModel(ABC):
    def __init__(self, settings: Settings):
        """
        Base class
        """
        # Settings
        self.settings = settings
        self.statistics = None
        self.loading = None

    def calculate_probability_loading(
        self,
        result_variable: str,
        levels: np.ndarray,
        model_uncertainty: bool = True,
        split_input_variables: list = [],
        given: list = [],
    ):
        """
        Determine the probability that the load (water level, hbn, qov, etc.) falls within a range.

        The probabilities are divided among load levels and any 'split variables'.

        It is possible to specify a variable as 'given'. The probability is then determined conditionally on
        this variable. This is often the slow stochastic variable, so that the waveshape can be processed later.

        The model uncertainty in the water level can be processed. It is then integrated. Note that
        independence between different block durations is assumed.

        Parameters
        ----------
        result_variable: str
            Load variable to be split.
        levels: np.ndarray
            Array with values of the load for which the probability is determined.
        model_uncertainty: bool
            Whether to integrate model uncertainty. Only applicable for water level 'h'.
        split_input_variables: list
            Variables to be split, in addition to the load variable.
        given: list
            List of variables on which the probability is conditionally determined.
        """
        # Statistics
        loading = self.get_loading()
        statistics = self.get_statistics()

        # Init the array
        shp = [len(levels) + 1]
        for var in split_input_variables:
            if var in list(statistics.stochastics_discrete.keys()):
                shp.extend([len(statistics.stochastics_discrete[var])])
            elif var in list(statistics.stochastics_fast.keys()):
                shp.extend([len(statistics.stochastics_fast[var])])
            elif var in list(statistics.stochastics_slow.keys()):
                shp.extend([len(statistics.stochastics_slow[var])])
        probability = np.zeros(shp)

        # Seperate the discrete stochastics from the continuous stochastics (slow and fast)
        contvars = [
            var
            for var in split_input_variables
            if var not in statistics.stochastics_discrete
        ]

        # Loop through all loading models (per r, k)
        for (direction, closing_situation), _ in loading.iter_models():
            # Get the IDs of the wind direction and closing situation
            ir = loading.r.index(direction)
            ik = loading.k.index(closing_situation)

            # Calculate the probability of the loading
            probability_loading = self.split_load_model(
                direction=direction,
                closing_situation=closing_situation,
                result_variable=result_variable,
                levels=levels,
                split_input_variables=contvars,
                given=given,
            )

            # If the combination of r, k is non-existent
            if np.sum(probability_loading) == 0.0:
                continue

            # Process model uncertainty
            if model_uncertainty:
                # Check
                if result_variable not in ["h", "hs", "tspec", "tp"]:
                    raise NotImplementedError(
                        "[ERROR] Processing model uncertainties is only possible for database result variables."
                    )

                # Bepaal grenzen
                boundaries = ProbabilityFunctions.calculate_boundaries(levels)

                # Verwerk de onzekerheid gegeven de sluitsituatie
                probability_loading = (
                    statistics.get_model_uncertainties().process_model_uncertainty(
                        closing_situation=closing_situation,
                        result_variable=result_variable,
                        levels=boundaries,
                        exceedance_probability=probability_loading,
                        haxis=0,
                    )
                )

            # Create an index to accumulate the probabilities.
            # If the discrete variables are included in the splitting, assign a specific position here.
            # Otherwise, they are added to the rest of the probabilities.
            idx = [slice(None)] * len(shp)
            if "r" in split_input_variables:
                idx[split_input_variables.index("r") + 1] = ir
            if "k" in split_input_variables:
                idx[split_input_variables.index("k") + 1] = ik
            probability[tuple(idx)] += probability_loading

        return probability

    def split_load_model(
        self,
        direction: float,
        closing_situation: int,
        result_variable: str,
        levels: np.ndarray,
        split_input_variables: list = [],
        given: list = [],
    ) -> np.ndarray:
        """
        Integrates the loads over the given levels and variables to split over.

        The probabilities are derived from the statistics based on the input variables in the load model.

        For the load type, any of the result variables can be chosen.

        P(H in [h1, h2], split_vars | given)

        Parameters
        ----------
        direction : float
        closing_situation : int
        result_variable : str
            Result variable
        levels : np.ndarray
            Levels for splitting the load.
        split_input_variables : list
            Variables to split over in addition to the load variable (default : []).
        given : list
            Conditional variables P(X|given) (default : []).

        Returns
        -------
        np.ndarray
            Probability per load variable and splitting variable.
        """
        # Statistics
        loading = self.get_loading()
        statistics = self.get_statistics()

        # Make a copy of the loading model for (r, k) to prevent overwriting the one assigned to this object
        loading_model = loading.model[(direction, closing_situation)]

        # Refine the loading onto the (finer) grid used in the statistics
        comb_fast_slow = {**statistics.stochastics_fast, **statistics.stochastics_slow}
        comb_fast_slow = {
            key: comb_fast_slow[key] for key in loading_model.input_variables
        }
        refined_load = loading_model.refine(result_variable, comb_fast_slow)

        # Calculate the probability of these loading combinations
        loading_probability = statistics.calculate_probability(
            loading_model.direction, loading_model.closing_situation, given=given
        )

        # Check the dimension from the loading and probability arrays
        assert refined_load.shape == tuple(
            loading_probability.shape[: refined_load.ndim]
        )
        if refined_load.size != loading_probability.size:
            # When the dimensions of the interpolated load (refined_load) and the probabilities do not match,
            # this must be due to an additional dimension in the probabilities where the load is equal.
            # The assumption is that this variable is dependent on the given conditions.
            loading_probability = loading_probability.reshape(
                (refined_load.size,) + loading_probability.shape[refined_load.ndim :]
            )
            extra_kansvar = given[:]

        else:
            loading_probability = loading_probability.ravel()
            extra_kansvar = []

        # Determine the dimensions of the variables to split over
        var_sizes = [
            len({**statistics.stochastics_fast, **statistics.stochastics_slow}[var])
            for var in split_input_variables
        ]

        # Create an array to allocate probabilities, this is an array for the load variable,
        # plus the determined dimensions
        probability = np.zeros((len(levels) + 1, *var_sizes))

        # Split the load into level classes
        digitized = np.digitize(refined_load.ravel(), levels)

        # Create an array with ones, which will be used to place the probability
        # in the correct position in the array by multiplying [1, 2, 3, ...] along the relevant axis,
        # and then raveling (flattening into 1D)
        ones = np.ones(refined_load.shape, dtype=digitized.dtype)
        base_shape = [1] * ones.ndim

        # Create a list of index arrays, with the load class as the first element
        idxs = [digitized]

        # Loop through all other variables
        for var_size, var in zip(var_sizes, split_input_variables):
            # Check if the variable exists
            if (var not in list(comb_fast_slow)) and (var not in extra_kansvar):
                raise ValueError(
                    f"'{var}' is not in list ({', '.join(list(comb_fast_slow))})."
                )
            if var in extra_kansvar:
                continue

            # Create the dimensions for assigning the arange
            shape = base_shape[:]
            shape[list(comb_fast_slow).index(var)] = -1

            # Determine the splitting positions, reshape and add to the list
            idxs.append((ones * np.arange(var_size).reshape(tuple(shape))).ravel())

        # Sum all probabilities into the relevant bin (based on the levels)
        if len(extra_kansvar) > 0:
            for i, idx in enumerate(zip(*idxs)):
                probability[idx] += loading_probability[i]
        else:
            np.add.at(probability, tuple(idxs), loading_probability)

        # Return the probabilities
        return probability

    def get_statistics(self) -> Statistics:
        """
        Return the Statistics object

        Returns
        -------
        Statistics
            The Statistics object for this Location
        """
        return self.statistics

    def get_loading(self) -> Loading:
        """
        Return the Loading object

        Returns
        -------
        Loading
            The Loading object for this Location
        """
        return self.loading



import numpy as np

from ..statistics import Statistics
from ....location import Location


class StatisticsWaveOvertopping(Statistics):
    def __init__(
        self, location: Location, water_levels: np.ndarray, probability: np.ndarray
    ):
        """
        Init the Statistics class for the Eastern Scheldt

        Parameters
        ----------
        settings : Settings
            The Settings object
        """
        # Inherit initialisation method from parent
        super().__init__(location.get_settings())

        # Init [wlev, u, xtr1, xtr2, ..., r]
        self.probability = probability

        #  Discrete, slow, fast stochatics
        self.stochastics_discrete = {
            "r": location.get_model().get_statistics().stochastics_discrete["r"]
        }
        self.stochastics_fast = {
            "wlev": water_levels,
            "u": location.get_model().get_statistics().stochastics_fast["u"],
        }
        self.stochastics_slow = location.get_model().get_statistics().stochastics_slow

        # Discretisation
        discretisation = {
            **self.stochastics_fast,
            **self.stochastics_slow,
            **self.stochastics_discrete,
        }
        self.dimensions = list(discretisation.keys())

        # Check dimensions
        for i, dim in enumerate(self.dimensions):
            if probability.shape[i] != len(discretisation[dim]):
                raise ValueError(
                    f"kansen.shape[i] {probability.shape[i]} != len(discretisatie[{dim}]) ({len(discretisation[dim])})."
                )
            setattr(self, f"n{dim}", len(discretisation[dim]))

    def calculate_probability(
        self, wind_direction: float, closing_situation: int = 1, given: list = []
    ) -> np.ndarray:
        """
        Return the probability of wlev, u, xtr1, xtr2 given the wind direction.

        Parameters
        ----------
        wind_direction : float
            The wind direction.
        closing_situation : int, optional
            The closing situation. Does not have any effect for
            StatisticsWaveOvertopping (default is 0)
        given : list, optional
            Given variables. Does not have any effect for
            StatisticsWaveOvertopping (default is 0)
        """
        # Obtain the wind direction id
        ir = list(self.stochastics_discrete["r"]).index(wind_direction)

        # Return the probability for this wind direction
        return np.take(self.probability, indices=ir, axis=self.dimensions.index("r"))



from .model.water_system import WaterSystem
from .profile.profile import Profile
from .settings.settings import Settings


class Location:
    """
    This is a class for one HRDLocation. This class contains the Settings,
    the WaterSystem mode (which is the model describing the statistics and
    loading) and a Profile. Which is a schematisation of the cross-section.
    """

    settings: Settings
    model: WaterSystem
    profile: Profile

    def __init__(self, settings: Settings, profile: Profile = None):
        """
        Initialize a Location object.

        Parameters
        ----------
        settings : Settings
            An instance of the Settings class that contains location-specific
            settings.
        profile : Profile
            An instance of the Profile class that provides a schematization of
            the cross-section.
        """
        self.settings = settings
        self.model = WaterSystem(settings)
        self.profile = profile

    def get_settings(self) -> Settings:
        """
        Return the Settings object.

        Returns
        -------
        Settings
            The Settings object for this Location
        """
        return self.settings

    def set_settings(self, settings: Settings) -> None:
        """
        Set the Settings for this location.

        Parameters
        ----------
        settings : Settings
            Settings object
        """
        # Check if the HRDLocation is still the same
        if self.settings.location != settings.location:
            raise ValueError(
                f"[ERROR] Cannot apply settings for location {settings.location} to location {self.settings.location}."
            )

        # Settings
        self.settings = settings

        # Statistics
        self.model = WaterSystem(settings)

    def get_model(self) -> WaterSystem:
        """
        Returns the WaterSystem model.

        Returns
        -------
        WaterSystem
            The WaterSystem model
        """
        return self.model

    def get_profile(self) -> Profile:
        """
        Return the Profile object

        Returns
        -------
        Loading
            The Loading object for this Location
        """
        return self.profile

    def has_profile(self) -> bool:
        """
        Returns whether of not a profile is assigned.

        Returns
        -------
        bool
            Profile assigned to Location
        """
        return isinstance(self.profile, Profile)

    def set_profile(self, profile: Profile = None) -> None:
        """
        Set the Profile for this location. By default, profile is none. Which
        deletes the profile from this location.

        Parameters
        ----------
        profile : Profile
            Profile object (default: None)
        """
        self.profile = profile

    def remove_profile(self) -> None:
        """
        Remove the Profile from the Location class.
        """
        self.profile = None



import numpy as np
import os

from ctypes import (
    ARRAY,
    CDLL,
    POINTER,
    Structure,
    c_bool,
    c_char,
    c_double,
    c_int,
    byref,
    create_string_buffer,
)


class ProfileLoading:
    """
    Profile loading object to calculate hbn, runup, overtopping and overflow
    discharge.
    """

    # Model factors
    MODEL_FACTORS = {
        "FactorDeterminationQbFn": 2.3,
        "FactorDeterminationQbFb": 4.3,
        "M_z2": 1.07,
        "Fshallow": 0.67778,
        "ComputedOvertopping": 1.0,
        "CriticalOvertopping": 1.0,
        "RelaxationFactor": 1.0,
        "ReductionFactorForeshore": 0.5,
    }

    def __init__(self, profile, settings: dict = None):
        """
        Constructor of the ProfileLoading object

        Parameters
        ----------
        profile : Profile
            The Profile object
        settings : dict
            Model uncertainty settings, optional (default = None)
            Parameters: FactorDeterminationQbFn, FactorDeterminationQbFb, M_z2,
            Fshallow, ComputedOvertopping, CriticalOvertopping,
            RelaxationFactor, ReductionFactorForeshore
        """
        # Update model factors if defined
        if isinstance(settings, dict):
            if all(k in list(self.MODEL_FACTORS.keys()) for k in list(settings.keys())):
                self.MODEL_FACTORS.update(settings)
            else:
                raise ValueError(
                    "[ERROR] Settings dictionary contains unknown keys (check uppercase?)."
                )

        # Path to the library
        lib_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "lib"))

        # Load RTO-libary (VTV v17.1.1.0) (used for runup / overtopping when the water level is below crest level)
        self.rto_library = CDLL(os.path.join(lib_path, "dllDikesOvertopping.dll"))

        # Load COO-library (used for overflow) (Note: DiKErnel does not include this .dll)
        self.coo_library = CDLL(os.path.join(lib_path, "CombOverloopOverslag64.dll"))

        # Modelfactors
        factors = np.array(list(self.MODEL_FACTORS.items()))[:, 1].astype(float)
        self.modelfactors_rto = (c_double * 8)(*factors)
        self.modelfactors_coo = c_double(factors[3])
        self.output = (c_double * 2)(0.0, 0.0)
        self.qo = c_double()
        self.niveau = c_double()
        self.load_rto = c_double * 4
        self.load_coo = OvertoppingLoad
        self.succes = c_bool()
        self.message = create_string_buffer(b"", 512)

        # Set profile
        self.profile = profile

    def set_profile(self, profile):
        """
        Set a profile

        Parameters
        ----------
        Profile : profile
            A Profile object
        """
        self.profile = profile

    def calculate_discharge(
        self,
        water_level: float,
        significant_wave_height: float,
        spectral_wave_period: float,
        wave_direction: float,
    ) -> float:
        """
        Calculate overtopping for loaded profile and given load combination.

        Parameters
        ----------
        water_level : float
            Water level
        significant_wave_height : float
            Significant wave height
        spectral_wave_period : float
            Spectral wave period
        wave_direction : float
            Wave direction

        Returns
        -------
        float
            The overtopping discharge
        """
        # If the waterlevel is lower or equal to the crest level, use rto
        if water_level <= self.profile.dike_crest_level:
            # Create instance of load structure and calculate the overtopping discharge
            load = self.load_rto(
                water_level,
                significant_wave_height,
                spectral_wave_period,
                wave_direction,
            )
            self.__calculate_discharge_rto(load)
            qov = self.output[0]

        # If the water is above the crestlevel use coo
        else:
            # Calculate overtopping at h = crest level
            load = self.load_rto(
                self.profile.dike_crest_level,
                significant_wave_height,
                spectral_wave_period,
                wave_direction,
            )
            self.__calculate_discharge_rto(load)
            qov_ot = self.output[0]

            # Calculate overflow
            load = self.load_coo(
                water_level,
                significant_wave_height,
                spectral_wave_period,
                wave_direction,
            )
            self.__calculate_discharge_coo(load)
            qov_ov = self.qo.value

            # Take the value from overflow only if it is larger than overtopping (to create a smooth function)
            if qov_ov > qov_ot:
                qov = qov_ov
            else:
                qov = qov_ot

        # Catch errors
        if not self.succes:
            raise ValueError(
                self.message.value.decode().strip()
                + f" (Load: h={water_level}, Hs={significant_wave_height}, Tm-1,0={spectral_wave_period}, wdir={wave_direction})"
            )

        # Return the overtopping discharge
        return qov

    def calculate_runup(
        self,
        water_level: float,
        significant_wave_height: float,
        spectral_wave_period: float,
        wave_direction: float,
    ) -> float:
        """
        Calculate run up (z2%) for loaded profile and given load combination.

        Parameters
        ----------
        water_level : float
            Water level
        significant_wave_height : float
            Significant wave height
        spectral_wave_period : float
            Spectral wave period
        wave_direction : float
            Wave direction

        Returns
        -------
        float
            The run up height (z2%)
        """
        # If the waterlevel is lower or equal to the crest level
        if water_level <= self.profile.dike_crest_level:
            # Calculate the run up
            load = self.load_rto(
                water_level,
                significant_wave_height,
                spectral_wave_period,
                wave_direction,
            )
            self.__calculate_discharge_rto(load)
            ru2p = self.output[1] + water_level

        # Otherwise give an error
        else:
            raise ValueError("Water level exceeds crest level.")

        # Catch errors
        if not self.succes:
            raise ValueError(
                self.message.value.decode().strip()
                + f" (Load: h={water_level}, Hs={significant_wave_height}, Tm-1,0={spectral_wave_period}, wdir={wave_direction})"
            )

        # Return the runup discharge
        return ru2p

    def calculate_crest_level(
        self,
        q_overtopping: float,
        water_level: float,
        significant_wave_height: float,
        spectral_wave_period: float,
        wave_direction: float,
    ) -> float:
        """
        Calculate the crest level for a discharge q_overtopping for loaded profile and given load combination.

        Parameters
        ----------
        q_overtopping : float
            Critical overtopping discharge
        water_level : float
            Water level
        significant_wave_height : float
            Significant wave height
        spectral_wave_period : float
            Spectral wave period
        wave_direction : float
            Wave direction

        Returns
        -------
        float
            The crest level
        """
        # Create instance of load structure
        self.qcr = c_double(q_overtopping)

        # If the waterlevel is lower or equal to the crest level, use rto
        if wave_direction > 360:
            if (wave_direction - 360) < 10e-4:
                wave_direction = 0
        load = self.load_rto(
            water_level, significant_wave_height, spectral_wave_period, wave_direction
        )
        self.__calculate_crest_level_rto(load)

        # Catch errors
        if not self.succes:
            raise ValueError(self.message.value.decode().strip())

        # If the calculated crest level is equal to the water level
        if self.niveau.value == water_level:
            load = self.load_coo(
                water_level,
                significant_wave_height,
                spectral_wave_period,
                wave_direction,
            )
            self.__calculate_crest_level_coo(load)

        # Catch errors
        if not self.succes:
            raise ValueError(self.message.value.decode().strip())

        return self.niveau.value

    def __calculate_discharge_rto(self, load):
        """
        Function to communicate with the dllDikesOvertopping.dll
        """
        self.arr_ctype = c_double * len(self.profile.dike_x_coordinates)
        self.__set_argtypes()
        self.rto_library.calculateQoJ(
            load,
            byref(self.arr_ctype(*self.profile.dike_x_coordinates)),
            byref(self.arr_ctype(*self.profile.dike_y_coordinates)),
            byref(self.arr_ctype(*self.profile.dike_roughness)),
            byref(c_double(self.profile.dike_orientation)),
            byref(c_int(len(self.profile.dike_x_coordinates))),
            byref(c_double(self.profile.dike_crest_level)),
            byref(self.modelfactors_rto),
            byref(self.output),
            byref(self.succes),
            byref(self.message),
        )

    def __calculate_crest_level_rto(self, load):
        """
        Function to communicate with the dllDikesOvertopping.dll
        """
        self.arr_ctype = c_double * len(self.profile.dike_x_coordinates)
        self.__set_argtypes()
        self.rto_library.omkeerVariantJ(
            load,
            byref(self.arr_ctype(*self.profile.dike_x_coordinates)),
            byref(self.arr_ctype(*self.profile.dike_y_coordinates)),
            byref(self.arr_ctype(*self.profile.dike_roughness)),
            byref(c_double(self.profile.dike_orientation)),
            byref(c_int(len(self.profile.dike_x_coordinates))),
            byref(self.qcr),
            byref(self.niveau),
            byref(self.modelfactors_rto),
            byref(self.output),
            byref(self.succes),
            byref(self.message),
        )

    def __calculate_discharge_coo(self, load):
        """
        Function to communicate with the CombOverloopOverslag64.dll
        """
        self.arr_ctype = c_double * len(self.profile.dike_x_coordinates)
        self.__set_argtypes()
        self.coo_library.CalculateDischarge(
            byref(c_double(self.profile.dike_orientation)),
            byref(c_int(len(self.profile.dike_x_coordinates))),
            byref(self.arr_ctype(*self.profile.dike_x_coordinates)),
            byref(self.arr_ctype(*self.profile.dike_y_coordinates)),
            byref(self.arr_ctype(*self.profile.dike_roughness)),
            load,
            byref(self.modelfactors_coo),
            byref(c_double(self.profile.dike_crest_level)),
            byref(self.qo),
            byref(self.succes),
            byref(self.message),
        )

    def __calculate_crest_level_coo(self, load):
        """
        Function to communicate with the CombOverloopOverslag64.dll
        """
        self.arr_ctype = c_double * len(self.profile.dike_x_coordinates)
        self.__set_argtypes()
        self.coo_library.CalculateHeight(
            byref(c_double(self.profile.dike_orientation)),
            byref(c_int(len(self.profile.dike_x_coordinates))),
            byref(self.arr_ctype(*self.profile.dike_x_coordinates)),
            byref(self.arr_ctype(*self.profile.dike_y_coordinates)),
            byref(self.arr_ctype(*self.profile.dike_roughness)),
            load,
            byref(self.modelfactors_coo),
            byref(self.qcr),
            byref(self.niveau),
            byref(self.succes),
            byref(self.message),
        )

    def __set_argtypes(self):
        """
        Set argtypes for dll functions

        This function is called after the profile is loaded
        """
        argtypes = {
            "load_rto": POINTER(c_double * 4),
            "load_coo": POINTER(OvertoppingLoad),
            "xp": POINTER(self.arr_ctype),
            "yp": POINTER(self.arr_ctype),
            "rp": POINTER(self.arr_ctype),
            "dike_orientation": POINTER(c_double),
            "debiet": POINTER(c_double),
            "npoints": POINTER(c_int),
            "crest_level": POINTER(c_double),
            "modelfactors_rto": POINTER(c_double * 8),
            "modelfactors_coo": POINTER(c_double),
            "output": POINTER(c_double * 2),
            "succes": POINTER(c_bool),
            "errormessage": POINTER(ARRAY(c_char, 512)),
        }

        # Overtopping and runup
        self.rto_library.calculateQoJ.argtypes = [
            argtypes[name]
            for name in [
                "load_rto",
                "xp",
                "yp",
                "rp",
                "dike_orientation",
                "npoints",
                "crest_level",
                "modelfactors_rto",
                "output",
                "succes",
                "errormessage",
            ]
        ]

        # HBN overtopping and runup
        self.rto_library.omkeerVariantJ.argtypes = [
            argtypes[name]
            for name in [
                "load_rto",
                "xp",
                "yp",
                "rp",
                "dike_orientation",
                "npoints",
                "debiet",
                "crest_level",
                "modelfactors_rto",
                "output",
                "succes",
                "errormessage",
            ]
        ]

        # Overflow
        self.coo_library.CalculateDischarge.argtypes = [
            argtypes[name]
            for name in [
                "dike_orientation",
                "npoints",
                "xp",
                "yp",
                "rp",
                "load_coo",
                "modelfactors_coo",
                "crest_level",
                "debiet",
                "succes",
                "errormessage",
            ]
        ]

        # HBN overflow
        self.coo_library.CalculateHeight.argtypes = [
            argtypes[name]
            for name in [
                "dike_orientation",
                "npoints",
                "xp",
                "yp",
                "rp",
                "load_coo",
                "modelfactors_coo",
                "debiet",
                "crest_level",
                "succes",
                "errormessage",
            ]
        ]


class OvertoppingLoad(Structure):
    """
    OvertoppingLoad class

    Contains crest level, wave height, wave period and wave direction.
    """

    _fields_ = [
        ("h", c_double),
        ("hm0", c_double),
        ("tm_10", c_double),
        ("phi", c_double),
    ]
